import os
import sys
import traceback
from dataclasses import dataclass
from enum import Enum
from typing import Generator, List, Optional, Tuple

from colorama import Fore, Style
from termcolor import colored

from ward.diff import build_auto_diff
from ward.expect import ExpectationFailed
from ward.suite import Suite
from ward.test_result import TestOutcome, TestResult


def truncate(s: str, num_chars: int) -> str:
    suffix = "..." if len(s) > num_chars - 3 else ""
    return s[:num_chars] + suffix


class ExitCode(Enum):
    SUCCESS = 0
    TEST_FAILED = 1
    ERROR = 2


class TestResultWriterBase:

    def __init__(self, suite: Suite):
        self.suite = suite
        self.terminal_size = get_terminal_size()

    def output_all_test_results(
        self,
        test_results_gen: Generator[TestResult, None, None],
        time_to_collect: float,
        fail_limit: Optional[int] = None,
    ) -> List[TestResult]:
        all_results = []
        failed_test_results = []
        print(f"Ward collected {self.suite.num_tests} tests and {self.suite.num_fixtures} fixtures "
              f"in {time_to_collect:.2f} seconds.\n")
        for result in test_results_gen:
            self.output_single_test_result(result)
            sys.stdout.write(Style.RESET_ALL)
            all_results.append(result)
            if result.outcome == TestOutcome.FAIL:
                failed_test_results.append(result)

            if len(failed_test_results) == fail_limit:
                break

        print()
        self.output_test_run_post_failure_summary(test_results=all_results)
        for failure in failed_test_results:
            self.output_why_test_failed_header(failure)
            self.output_why_test_failed(failure)

        return all_results

    def output_single_test_result(self, test_result: TestResult):
        """Indicate whether a test passed, failed, was skipped etc."""
        raise NotImplementedError()

    def output_why_test_failed_header(self, test_result: TestResult):
        """
        Printed above the failing test output
        """
        raise NotImplementedError()

    def output_test_result_summary(self, test_results: List[TestResult], time_taken: float):
        raise NotImplementedError()

    def output_why_test_failed(self, test_result: TestResult):
        """
        Extended output shown for failing tests, may include further explanations,
        assertion error info, diffs, etc.
        """
        raise NotImplementedError()

    def output_test_run_post_failure_summary(self, test_results: List[TestResult]):
        raise NotImplementedError()


def lightblack(s: str) -> str:
    return f"{Fore.LIGHTBLACK_EX}{s}{Style.RESET_ALL}"


@dataclass
class TerminalSize:
    height: int
    width: int


def get_terminal_size() -> TerminalSize:
    for i in range(0, 3):
        try:
            cols, rows = os.get_terminal_size(i)
            return TerminalSize(height=rows, width=cols)
        except OSError:
            continue
    return TerminalSize(height=24, width=80)


class SimpleTestResultWrite(TestResultWriterBase):

    def output_single_test_result(self, test_result: TestResult):
        outcome_to_colour = {
            TestOutcome.PASS: "green",
            TestOutcome.SKIP: "blue",
            TestOutcome.FAIL: "red",
        }
        colour = outcome_to_colour[test_result.outcome]
        bg = f"on_{colour}"
        padded_outcome = f" {test_result.outcome.name} "
        mod_name = lightblack(f"{test_result.test.module.__name__}.")
        print(colored(padded_outcome, color='grey', on_color=bg),
              mod_name + test_result.test.name)

    def output_why_test_failed_header(self, test_result: TestResult):
        print(colored(" Failure", color="red", attrs=["bold"]), "in",
              colored(test_result.test.qualified_name, attrs=["bold"]))

    def output_why_test_failed(self, test_result: TestResult):
        truncation_chars = self.terminal_size.width - 24
        err = test_result.error
        if isinstance(err, ExpectationFailed):
            print(f"\n  Given {truncate(repr(err.history[0].this), num_chars=truncation_chars)}")

            for expect in err.history:
                if expect.success:
                    result_marker = f"[ {Fore.GREEN}✓{Style.RESET_ALL} ]{Fore.GREEN}"
                else:
                    result_marker = f"[ {Fore.RED}✗{Style.RESET_ALL} ]{Fore.RED}"

                if expect.op == "satisfies" and hasattr(expect.that, "__name__"):
                    expect_that = truncate(expect.that.__name__, num_chars=truncation_chars)
                else:
                    expect_that = truncate(repr(expect.that), num_chars=truncation_chars)
                print(
                    f"    {result_marker} it {expect.op} {expect_that}{Style.RESET_ALL}",
                )

            if err.history and err.history[-1].op == "equals":
                expect = err.history[-1]
                print(
                    f"\n  Showing diff of {colored('expected value', color='green')}"
                    f" vs {colored('actual value', color='red')}:\n")

                diff = build_auto_diff(expect.that, expect.this, width=truncation_chars)
                print(diff)
        else:
            trace = getattr(err, "__traceback__", "")
            if trace:
                trc = traceback.format_exception(None, err, trace)
                print("".join(trc))
            else:
                print(str(err))

        print(Style.RESET_ALL)

    def output_test_result_summary(self, test_results: List[TestResult], time_taken: float):
        num_passed, num_failed, num_skipped = self._get_num_passed_failed_skipped(test_results)
        print(self.generate_chart(num_passed=num_passed, num_failed=num_failed, num_skipped=num_skipped), "")

        if any(r.outcome == TestOutcome.FAIL for r in test_results):
            result = colored("FAILED", color='red', attrs=["bold"])
        else:
            result = colored("PASSED", color='green', attrs=["bold"])
        print(f"{result} in {time_taken:.2f} seconds [ "
              f"{colored(str(num_failed) + ' failed', color='red')}  "
              f"{colored(str(num_skipped) + ' skipped', color='blue')}  "
              f"{colored(str(num_passed) + ' passed', color='green')} ]")

    def generate_chart(self, num_passed, num_failed, num_skipped):
        pass_pct = num_passed / max(num_passed + num_failed + num_skipped, 1)
        fail_pct = num_failed / max(num_passed + num_failed + num_skipped, 1)
        skip_pct = 1.0 - pass_pct - fail_pct

        num_green_bars = int(pass_pct * self.terminal_size.width)
        num_red_bars = int(fail_pct * self.terminal_size.width)
        num_blue_bars = int(skip_pct * self.terminal_size.width)

        # Rounding to integers could leave us a few bars short
        num_bars_remaining = self.terminal_size.width - num_green_bars - num_red_bars - num_blue_bars
        if num_bars_remaining and num_green_bars:
            num_green_bars += 1
            num_bars_remaining -= 1

        if num_bars_remaining and num_red_bars:
            num_red_bars += 1
            num_bars_remaining -= 1

        if num_bars_remaining and num_blue_bars:
            num_blue_bars += 1
            num_bars_remaining -= 1

        assert num_bars_remaining == 0

        if self.terminal_size.width - num_green_bars - num_red_bars == 1:
            num_green_bars += 1

        return (colored("F" * num_red_bars, color="red", on_color="on_red") +
                colored("s" * num_blue_bars, color="blue", on_color="on_blue") +
                colored("." * num_green_bars, color="green", on_color="on_green"))

    def output_test_run_post_failure_summary(self, test_results: List[TestResult]):
        pass

    def _get_num_passed_failed_skipped(self, test_results: List[TestResult]) -> Tuple[int, int, int]:
        num_passed = len([r for r in test_results if r.outcome == TestOutcome.PASS])
        num_failed = len([r for r in test_results if r.outcome == TestOutcome.FAIL])
        num_skipped = len([r for r in test_results if r.outcome == TestOutcome.SKIP])

        return num_passed, num_failed, num_skipped



import os
import platform
import sys
from timeit import default_timer

import click
from colorama import init

from ward.collect import get_info_for_modules, get_tests_in_modules, load_modules
from ward.fixtures import fixture_registry
from ward.suite import Suite
from ward.terminal import ExitCode, SimpleTestResultWrite
from ward.test_result import TestOutcome

init()

if platform.system() == "Windows":
    os.system('color')


@click.command()
@click.option(
    "-p", "--path", default=".", type=click.Path(exists=True), help="Path to tests."
)
@click.option(
    "-f", "--filter", help="Only run tests whose names contain the filter argument as a substring."
)
@click.option(
    "--fail-limit", type=int, help="The number of failures to cancel the run after."
)
def run(path, filter, fail_limit):
    start_run = default_timer()

    mod_infos = get_info_for_modules(path)
    modules = list(load_modules(mod_infos))
    tests = list(get_tests_in_modules(modules, filter=filter))
    time_to_collect = default_timer() - start_run

    suite = Suite(tests=tests, fixture_registry=fixture_registry)
    test_results = suite.generate_test_runs()

    writer = SimpleTestResultWrite(suite=suite)
    results = writer.output_all_test_results(
        test_results,
        time_to_collect=time_to_collect,
        fail_limit=fail_limit,
    )
    time_taken = default_timer() - start_run
    writer.output_test_result_summary(results, time_taken)

    if any(r.outcome == TestOutcome.FAIL for r in results):
        exit_code = ExitCode.TEST_FAILED
    else:
        exit_code = ExitCode.SUCCESS

    sys.exit(exit_code.value)



from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional

from ward.test import Test


class TestOutcome(Enum):
    PASS = auto()
    FAIL = auto()
    SKIP = auto()
    XFAIL = auto()


@dataclass
class TestResult:
    test: Test
    outcome: TestOutcome
    error: Optional[Exception]
    message: str



import importlib
import importlib.util
import os
import pkgutil
from importlib._bootstrap import ModuleSpec
from importlib._bootstrap_external import FileFinder
from typing import Any, Generator, Iterable

from ward.test import Test, WardMarker


def is_test_module(module: pkgutil.ModuleInfo) -> bool:
    return module.name.startswith("test_")


def get_info_for_modules(path: str) -> Generator[pkgutil.ModuleInfo, None, None]:
    # Check for modules at the root of the specified path
    for module in pkgutil.iter_modules([path]):
        yield module

    # Now check for modules in every subdirectory
    for root, dirs, _ in os.walk(path):
        for dir_name in dirs:
            dir_path = os.path.join(root, dir_name)
            for module in pkgutil.iter_modules([dir_path]):
                if is_test_module(module):
                    yield module


def load_modules(modules: Iterable[pkgutil.ModuleInfo]) -> Generator[Any, None, None]:
    for m in modules:
        file_finder: FileFinder = m.module_finder
        spec: ModuleSpec = file_finder.find_spec(m.name)
        mod = importlib.util.module_from_spec(spec)
        if mod.__name__.startswith("test_"):
            spec.loader.exec_module(mod)
            yield mod


def get_tests_in_modules(modules: Iterable[Any], filter: str = "") -> Generator[Test, None, None]:
    for mod in modules:
        for item in dir(mod):
            if item.startswith("test_"):
                test_name = item
                test_fn = getattr(mod, test_name)
                marker = getattr(test_fn, "ward_marker", WardMarker.NONE)
                if test_fn:
                    test = Test(
                        fn=test_fn,
                        module=mod,
                        marker=marker,
                    )

                    # Yield tests if there's no filter, or if the filter matches
                    if not filter:
                        yield test
                    elif filter in test.qualified_name:
                        yield test



from dataclasses import dataclass
from typing import Generator, List

from ward.fixtures import FixtureExecutionError, FixtureRegistry
from ward.test import Test, WardMarker
from ward.test_result import TestResult, TestOutcome


@dataclass
class Suite:
    tests: List[Test]
    fixture_registry: FixtureRegistry

    @property
    def num_tests(self):
        return len(self.tests)

    @property
    def num_fixtures(self):
        return len(self.fixture_registry)

    def generate_test_runs(self) -> Generator[TestResult, None, None]:
        for test in self.tests:
            if test.marker == WardMarker.SKIP:
                yield TestResult(test, TestOutcome.SKIP, None, "")
                continue

            try:
                resolved_fixtures = test.resolve_args(self.fixture_registry)
            except FixtureExecutionError as e:
                yield TestResult(test, TestOutcome.FAIL, e, message="[Error] " + str(e))
                continue
            try:
                resolved_vals = {k: fix.resolved_val for (k, fix) in resolved_fixtures.items()}
                test(**resolved_vals)
                yield TestResult(test, TestOutcome.PASS, None, message="")
            except Exception as e:
                yield TestResult(test, TestOutcome.FAIL, e, message="")
            finally:
                for fixture in resolved_fixtures.values():
                    if fixture.is_generator_fixture:
                        try:
                            fixture.cleanup()
                        except (RuntimeError, StopIteration):
                            # In Python 3.7, a RuntimeError is raised if we fall off the end of a generator
                            # (instead of a StopIteration)
                            pass



import inspect
from dataclasses import dataclass
from enum import Enum, auto
from types import MappingProxyType, ModuleType
from typing import Callable, Dict

from ward.fixtures import Fixture, FixtureRegistry


class WardMarker(Enum):
    NONE = auto()
    SKIP = auto()
    XFAIL = auto()


def skip(func):
    func.ward_marker = WardMarker.SKIP
    return func


def xfail(func):
    func.ward_marker = WardMarker.XFAIL
    return func


@dataclass
class Test:
    fn: Callable
    module: ModuleType
    marker: WardMarker = WardMarker.NONE

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    @property
    def name(self):
        return self.fn.__name__

    @property
    def qualified_name(self):
        return f"{self.module.__name__}.{self.name}"

    def deps(self) -> MappingProxyType:
        return inspect.signature(self.fn).parameters

    def has_deps(self) -> bool:
        return len(self.deps()) > 0

    def resolve_args(self, fixture_registry: FixtureRegistry) -> Dict[str, Fixture]:
        """Resolve fixture that has been injected into this test"""
        if not self.has_deps():
            return {}

        # Construct a dict of kwargs to pass into the test when it's called
        resolved_args = {}
        for fixture_name in self.deps():
            fixture = fixture_registry[fixture_name]
            resolved_arg = fixture.resolve(fixture_registry)
            resolved_args[fixture_name] = resolved_arg

        return resolved_args



import functools
import math
from dataclasses import dataclass
from typing import Type, Any, List, Callable, Dict, Tuple


class raises:
    def __init__(self, expected_ex_type: Type[Exception]):
        self.expected_ex_type = expected_ex_type

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not self.expected_ex_type:
            raise AssertionError(
                f"Expected exception {self.expected_ex_type}, but {exc_type} was raised instead."
            )
        return True


@dataclass
class Expected:
    this: Any
    op: str
    that: Any
    op_args: Tuple
    op_kwargs: Dict
    success: bool = True


class ExpectationFailed(Exception):
    def __init__(self, message: str, history: List[Expected]):
        self.message = message
        self.history = history


def record_expect_in_history(func):
    @functools.wraps(func)
    def wrapped_func(self, that: Any, *args, **kwargs) -> "expect":
        rv = func(self, that, *args, **kwargs)
        if rv:
            self.history.append(
                Expected(this=self.this, op=func.__name__, that=that, success=True, op_args=args, op_kwargs=kwargs)
            )
            return self
        else:
            self.history.append(
                Expected(this=self.this, op=func.__name__, that=that, success=False, op_args=args, op_kwargs=kwargs)
            )
            raise ExpectationFailed(f"{func.__name__} expectation failed", self.history)

    return wrapped_func


class expect:
    def __init__(self, this: Any):
        self.this = this
        self.history: List[Expected] = []

    @record_expect_in_history
    def equals(self, that: Any):
        return self.this == that

    @record_expect_in_history
    def is_less_than(self, that: Any):
        return self.this < that

    @record_expect_in_history
    def is_greater_than(self, that: Any):
        return self.this > that

    @record_expect_in_history
    def contains(self, that: Any):
        return that in self.this

    @record_expect_in_history
    def has_length(self, length: int):
        return len(self.this) == length

    @record_expect_in_history
    def is_instance_of(self, type: Type):
        return isinstance(self.this, type)

    @record_expect_in_history
    def satisfies(self, predicate: Callable[["expect"], bool]):
        return predicate(self.this)

    @record_expect_in_history
    def is_(self, that: Any):
        return self.this is that

    @record_expect_in_history
    def approx(self, that: Any, epsilon: float):
        return math.isclose(self.this, that, abs_tol=epsilon)



import difflib
import pprint

from termcolor import colored


def build_auto_diff(lhs, rhs, width=60) -> str:
    """Determines the best type of diff to use based on the output"""

    lhs_repr = pprint.pformat(lhs, width=width)
    rhs_repr = pprint.pformat(rhs, width=width)
    # TODO: Right now, just stick to unified diff while deciding what to do
    # if "\n" in lhs_repr and "\n" in rhs_repr:
    #     diff = build_unified_diff(lhs_repr, rhs_repr)
    # else:
    #     diff = build_split_diff(lhs_repr, rhs_repr)

    return build_unified_diff(lhs_repr, rhs_repr)


def build_split_diff(lhs_repr, rhs_repr) -> str:
    lhs_out, rhs_out = "", ""

    matcher = difflib.SequenceMatcher(None, lhs_repr, rhs_repr)
    for op, i1, i2, j1, j2 in matcher.get_opcodes():

        lhs_substring_lines = lhs_repr[i1:i2].splitlines()
        rhs_substring_lines = rhs_repr[j1:j2].splitlines()

        for i, lhs_substring in enumerate(lhs_substring_lines):
            if op == "replace":
                lhs_out += colored(lhs_substring, color="green")
            elif op == "delete":
                lhs_out += colored(lhs_substring, color="green")
            elif op == "insert":
                lhs_out += lhs_substring
            elif op == "equal":
                lhs_out += lhs_substring

            if i != len(lhs_substring_lines) - 1:
                lhs_out += f"\n"

        for j, rhs_substring in enumerate(rhs_substring_lines):
            if op == "replace":
                rhs_out += colored(rhs_substring, color="red")
            elif op == "insert":
                rhs_out += colored(rhs_substring, color="red")
            elif op == "equal":
                rhs_out += rhs_substring

            if j != len(rhs_substring_lines) - 1:
                rhs_out += f"\n"

    # TODO: Clean up the line below
    return f"LHS: {lhs_out}\nRHS: {rhs_out}"


def build_unified_diff(lhs_repr, rhs_repr, margin_left=4) -> str:
    differ = difflib.Differ()
    lines_lhs = lhs_repr.splitlines()
    lines_rhs = rhs_repr.splitlines()
    diff = differ.compare(lines_lhs, lines_rhs)

    output = []
    for line in diff:
        if line.startswith("- "):
            output.append(colored(line[2:], color="green"))
        elif line.startswith("+ "):
            output.append(colored(line[2:], color="red"))
        elif line.startswith("? "):
            # We can use this to find the index of change in
            # the line above if required in the future
            pass
        else:
            output.append(line[2:])

    return " " * margin_left + f"\n{' ' * margin_left}".join(output)



import inspect
from typing import Callable, Dict, Iterable


class TestSetupError(Exception):
    pass


class CollectionError(TestSetupError):
    pass


class FixtureExecutionError(Exception):
    pass


class Fixture:
    def __init__(self, key: str, fn: Callable, is_generator_fixture: bool = False):
        self.key = key
        self.fn = fn
        self.is_generator_fixture = is_generator_fixture
        self.gen = None
        self.resolved_val = None
        self.is_resolved = False

    def deps(self):
        return inspect.signature(self.fn).parameters

    def resolve(self, fix_registry) -> "Fixture":
        """Traverse the fixture tree to resolve the value of this fixture"""

        # If this fixture has no children, cache and return the resolved value
        if not self.deps():
            try:
                if inspect.isgeneratorfunction(self.fn):
                    self.gen = self.fn()
                    self.resolved_val = next(self.gen)
                else:
                    self.resolved_val = self.fn()
            except Exception as e:
                raise FixtureExecutionError(
                    f"Unable to execute fixture '{self.key}'"
                ) from e
            fix_registry.cache_fixture(self)
            return self

        # Otherwise, we have to find the child fixture vals, and call self
        children = self.deps()
        children_resolved = []
        for child in children:
            child_fixture = fix_registry[child].resolve(fix_registry)
            children_resolved.append(child_fixture)

        # We've resolved the values of all child fixtures
        try:
            child_resolved_vals = [child.resolved_val for child in children_resolved]
            if inspect.isgeneratorfunction(self.fn):
                self.gen = self.fn(*child_resolved_vals)
                self.resolved_val = next(self.gen)
            else:
                self.resolved_val = self.fn(*child_resolved_vals)
        except Exception as e:
            raise FixtureExecutionError(
                f"Unable to execute fixture '{self.key}'"
            ) from e

        fix_registry.cache_fixture(self)
        return self

    def cleanup(self):
        if self.is_generator_fixture:
            next(self.gen)


class FixtureRegistry:
    def __init__(self):
        self._fixtures: Dict[str, Fixture] = {}

    @property
    def decorator(self):
        def wrapper(func):
            name = func.__name__
            if name not in self._fixtures:
                self._fixtures[name] = Fixture(
                    key=name,
                    fn=func,
                    is_generator_fixture=inspect.isgeneratorfunction(func),
                )
            else:
                raise CollectionError(f"Multiple fixtures named '{func.__name__}'.")
            return func

        return wrapper

    def _get_fixture(self, fixture_name: str) -> Fixture:
        try:
            return self._fixtures[fixture_name]
        except KeyError:
            raise CollectionError(f"Couldn't find fixture '{fixture_name}'.")

    def cache_fixture(self, fixture: Fixture):
        """Update the fixture in the registry, for example, replace it with its resolved analogue"""
        # TODO: Caching can be used to implement fixture scoping,
        #  but currently resolved cached fixtures aren't used.
        self._fixtures[fixture.key] = fixture

    def cache_fixtures(self, fixtures: Iterable[Fixture]):
        for fixture in fixtures:
            self.cache_fixture(fixture)

    def __getitem__(self, item):
        return self._fixtures[item]

    def __len__(self):
        return len(self._fixtures)


fixture_registry = FixtureRegistry()
fixture = fixture_registry.decorator
